import pyruvate
import pytest
import requests
from multiprocessing import Process
from wsgiref.validate import validator
from time import sleep


HELLO_WORLD = b"<html><body><h1>Hello world!</h1></body></html>\n"


def simple_app(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/html')]
    start_response(status, response_headers)
    return [HELLO_WORLD]


@pytest.fixture
def one_worker(capsys):
    vldt = validator(simple_app)

    def srv():
        pyruvate.serve(vldt, '127.0.0.1:7878', 1)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


class TestPyruvate(object):

    def _pyruvate_serve(self, fixture):
        while True:
            tries = 0
            try:
                got = requests.get('http://localhost:7878/')
                break
            except ConnectionRefusedError:
                if tries == 2:
                    raise
                tries += 1
                sleep(1)
        assert got.ok
        assert got.status_code == 200
        assert got.content == b'<html><body><h1>Hello world!</h1></body></html>\n'  # noqa: E501
        assert got.headers['Content-type'] == 'text/html'

    def test_serve(self, one_worker):
        self._pyruvate_serve(one_worker)



import pyruvate
import pytest
import os
import socket
from multiprocessing import Process
from tempfile import mkstemp
from time import sleep


def filewrapper_app(environ, start_response):
    fh, fname = mkstemp()
    os.write(fh, b'Hello world!\n')
    os.close(fh)
    fw = environ.get('wsgi.file_wrapper')
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain'), ('Content-Length', 5)]
    start_response(status, response_headers)
    return fw(open(fname))


def honour_blocksize_app(environ, start_response):
    fh, fname = mkstemp()
    os.write(fh, b'Hello world!\n')
    os.close(fh)
    fw = environ.get('wsgi.file_wrapper')
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain'), ('Content-Length', 13)]
    start_response(status, response_headers)
    return fw(open(fname), blocksize=4)


@pytest.fixture
def one_worker_content_length(capsys):
    def srv():
        pyruvate.serve(
            filewrapper_app, '127.0.0.1:7878', 1, max_reuse_count=10)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


@pytest.fixture
def one_worker_blocksize(capsys):
    def srv():
        pyruvate.serve(
            honour_blocksize_app,
            '127.0.0.1:7878',
            1,
            max_reuse_count=10)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


class TestFileWrapper(object):

    def _honour_content_length_header_sendfile(self, fixture):
        # we cannot rely on higher level modules/libraries
        # because they will not read beyond Content-Length
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            tries = 0
            while True:
                try:
                    s.connect(('127.0.0.1', 7878))
                    break
                except ConnectionRefusedError:
                    if tries == 2:
                        raise
                    tries += 1
                    sleep(1)
            s.sendall(b'GET / HTTP/1.1\r\n\r\n')
            data = b''
            while True:
                chunk = s.recv(1024)
                data += chunk
                if len(data) >= 108:
                    break
            assert data == b'HTTP/1.1 200 OK\r\nContent-type: text/plain\r\nContent-Length: 5\r\nVia: pyruvate\r\nConnection: keep-alive\r\n\r\nHello'  # noqa: E501

    def _honour_blocksize_sendfile(self, fixture):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            tries = 0
            while True:
                try:
                    s.connect(('127.0.0.1', 7878))
                    break
                except ConnectionRefusedError:
                    if tries == 2:
                        raise
                    tries += 1
                    sleep(1)
            s.sendall(b'GET / HTTP/1.1\r\n\r\n')
            data = b''
            while True:
                chunk = s.recv(1024)
                data += chunk
                if len(data) >= 117:
                    break
            assert data == b'HTTP/1.1 200 OK\r\nContent-type: text/plain\r\nContent-Length: 13\r\nVia: pyruvate\r\nConnection: keep-alive\r\n\r\nHello world!\n'  # noqa: E501

    def test_honour_content_length_header_sendfile(
            self, one_worker_content_length):
        self._honour_content_length_header_sendfile(
                one_worker_content_length)

    def test_honour_blocksize_sendfile(
            self, one_worker_blocksize):
        self._honour_blocksize_sendfile(one_worker_blocksize)



import pyruvate
import pytest
import requests
import socket
from multiprocessing import Process
from time import sleep


def dummy_app(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return [b'OK']


def latin1_url_app(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return [bytes(environ['PATH_INFO'], encoding='latin-1')]


def long_header_app(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return [bytes(environ['HTTP_FOO'], encoding='ascii')]


@pytest.fixture
def dummy(capsys):
    def srv():
        pyruvate.serve(dummy_app, '127.0.0.1:7878', 1)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


@pytest.fixture
def latin1(capsys):
    def srv():
        pyruvate.serve(latin1_url_app, '127.0.0.1:7878', 1)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


@pytest.fixture
def long_header(capsys):
    def srv():
        pyruvate.serve(long_header_app, '127.0.0.1:7878', 1)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


@pytest.fixture
def content_length_wrong(capsys):
    def content_length_wrong_app(environ, start_response):
        status = '200 OK'
        response_headers = [
            ('Content-type', 'text/plain'),
            ('Content-Length', '42')]
        start_response(status, response_headers)
        return [b'Hello world!\n']
    def srv():
        pyruvate.serve(content_length_wrong_app, '127.0.0.1:7878', 1)
    p = Process(target=srv)
    yield p.start()
    p.terminate()


class TestHeaders(object):

    def _get_long_header(self):
        long_header = b'Foo42' * 65353
        tries = 0
        while True:
            try:
                got = requests.get(
                    'http://localhost:7878/', headers={'Foo': long_header})
                break
            except ConnectionRefusedError:
                if tries == 2:
                    raise
                tries += 1
                sleep(1)
        assert got.ok
        assert got.status_code == 200
        assert got.content == long_header  # noqa: E501
        assert got.headers['Content-type'] == 'text/plain'

    def _get_latin1_url(self):
        got = requests.get('http://localhost:7878/f%C3%A4%C3%A4')
        assert got.ok
        assert got.status_code == 200
        assert got.content == b'/f\xc3\xa4\xc3\xa4'

    def test_long_header(self, long_header):
        self._get_long_header()

    def test_latin1_url(self, latin1):
        self._get_latin1_url()

    def _expect_continue(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            tries = 0
            while True:
                try:
                    s.connect(('127.0.0.1', 7878))
                    break
                except ConnectionRefusedError:
                    if tries == 2:
                        raise
                    tries += 1
                    sleep(1)
            s.sendall(b'POST / HTTP/1.1\r\nContent-Length: 1603\r\nContent-Type: application/x-www-form-urlencoded\r\nExpect: 100-continue\r\n\r\n')  # noqa: E501
            data = b''
            while True:
                chunk = s.recv(1024)
                data += chunk
                if len(data) >= 16:
                    break
            assert data == b'HTTP/1.1 100 Continue\r\n\r\n'  # noqa: E501
            s.sendall(b'B' * 1603)
            data = b''
            while True:
                chunk = s.recv(1024)
                data += chunk
                if len(data) >= 16:
                    break
            assert data == b'HTTP/1.1 200 OK\r\nContent-type: text/plain\r\nVia: pyruvate\r\nConnection: close\r\n\r\nOK'  # noqa: E501

    def test_expect_continue(self, dummy):
        self._expect_continue()

    def test_content_length_wrong(self, content_length_wrong):
        try:
            tries = 0
            while True:
                try:
                    requests.get('http://localhost:7878/')
                    break
                except ConnectionRefusedError:
                    if tries == 2:
                        raise
                    tries += 1
                    sleep(1)
        except requests.exceptions.ChunkedEncodingError as e:
            incomplete_read = e.args[0].args[1]
            assert incomplete_read.expected == 29
            assert incomplete_read.partial == 13



import aiohttp
import asyncio
import logging
import logging.handlers
import pyruvate
import pytest
import queue
import multiprocessing
import multiprocessing.connection
from time import sleep


def dummy_app(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    sleep(.3)
    start_response(status, response_headers)
    return [b'OK']


queued_log = multiprocessing.Queue()


@pytest.fixture
def rq_monitor_fixture(caplog):
    def srv(logs):
        h = logging.handlers.QueueHandler(logs)
        root = logging.getLogger()
        root.addHandler(h)
        root.setLevel(logging.WARN)
        pyruvate.serve(
            dummy_app, '127.0.0.1:7878', 1, qmon_warn_threshold=2)
    p = multiprocessing.Process(target=srv, args=(queued_log, ))
    yield p.start()
    p.join(2)
    p.terminate()


class TestRequestQueueMonitor(object):

    async def _get_response(self, session, url):
        async with session.get('http://localhost:7878/') as resp:
            assert resp.ok
            assert resp.status == 200
            assert resp.headers['Content-type'] == 'text/plain'
            got = await resp.text()
            assert got == 'OK'

    @pytest.mark.asyncio
    async def test_queue_monitor(self, rq_monitor_fixture):
        """ Produce concurrent requests
        """
        url = 'http://localhost:7878/'
        async with aiohttp.ClientSession() as session:
            resps = []
            for rd in range(5):
                resps.append(
                    asyncio.ensure_future(self._get_response(session, url)))
            await asyncio.gather(*resps)
        record_count = 0
        while True:
            try:
                record = queued_log.get_nowait()
                record_count += 1
                assert record.levelno == logging.WARN
                assert 'requests in queue' in record.message
                assert record.name == 'pyruvate'
            except queue.Empty:
                break
        assert record_count > 0
